
from bergen.auths.types import HerreConfig
from oauthlib.oauth2.rfc6749.errors import InvalidClientError
from bergen.auths.base import AuthError, BaseAuthBackend
from oauthlib.oauth2.rfc6749.clients.backend_application import BackendApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session 
from bergen.enums import ClientType
import logging
import time
from bergen.console import console
logger = logging.getLogger(__name__)


class ArnheimBackendOauth(BaseAuthBackend):

    def __init__(self, config: HerreConfig, **kwargs) -> None:
        super().__init__(config , **kwargs)  

    def fetchToken(self, loop=None) -> str:
        # Connecting
        
        with console.status("[bold green]Authenticating..."):
            auth_client = BackendApplicationClient(client_id=self.client_id, scope=self.scope)
            oauth_session = OAuth2Session(client=auth_client, scope=self.scope)

            token = oauth_session.fetch_token(token_url=self.token_url, client_id=self.client_id,
            client_secret=self.client_secret, verify=True)

        return token
                




from bergen.messages.base import MessageModel
from bergen.messages.utils import expandToMessage
import json
from bergen.entertainer.base import BaseEntertainer
import logging
import asyncio
import websockets
from bergen.console import console
import asyncio
try:
    from asyncio import create_task
except ImportError:
    #python 3.6 fix
    create_task = asyncio.ensure_future


logger = logging.getLogger()





class WebsocketEntertainer(BaseEntertainer):
    ''' Is a mixin for Our Bergen '''

    def __init__(self, host= None, port= None, protocol=None, auto_reconnect=True, auth: dict= None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.websocket_host = host
        self.websocket_port = port
        self.websocket_protocol = protocol
        self.token = auth["token"]
        
        self.auto_reconnect= auto_reconnect
        self.allowed_retries = 2
        self.current_retries = 0

    async def connect(self):
        self.incoming_queue = asyncio.Queue()
        self.outgoing_queue = asyncio.Queue()
        self.tasks = []

        self.startup_task = create_task(self.startup())

    async def disconnect(self) -> str:
        for reference, task in self.assignments.items():
            if not task.done():
                logger.info(f"Cancelling Assignment {task}")
                task.cancel()

        if self.connection: await self.connection.close()

        if self.pending:
            for task in self.pending:
                task.cancel()

        logger.info("Entertainer disconnected")

        
    async def startup(self):
        try:
            await self.connect_websocket()
        except Exception as e:

            logger.error(f"Entertainer Connection failed {e}")
            self.current_retries += 1
            if self.current_retries < self.allowed_retries and self.auto_reconnect:
                sleeping_time = (self.current_retries + 1)
                logger.info(f"Retrying in {sleeping_time} seconds")
                await asyncio.sleep(sleeping_time)
                await self.startup()
            else:
                logger.error("No reconnecting attempt envisioned. Shutting Down!")
                return

        self.consumer_task = create_task(
            self.consumer()
        )

        self.producer_task = create_task(
            self.producer()
        )

        self.worker_task = create_task(
            self.workers()
        )

        done, self.pending = await asyncio.wait(
            [self.consumer_task, self.worker_task, self.producer_task],
            return_when=asyncio.FIRST_EXCEPTION
        )

        logger.error(f"Lost connection inbetween everything :( {[ task.exception() for task in done]}")
        logger.error(f'Reconnecting')

        for task in self.pending:
            task.cancel()

        if self.connection: await self.connection.close()
        self.current_retries = 0 # reset retries after one successfull connection
        await self.startup() # Attempt to ronnect again
        

    async def connect_websocket(self):

        uri = f"{self.websocket_protocol}://{self.websocket_host}:{self.websocket_port}/host/?token={self.token}"
        
        self.connection = await websockets.client.connect(uri)
        logger.info("Connecting as Entertainer")


    async def consumer(self):
        logger.warning(" [x] Awaiting Entertaining Calls")
        async for message in self.connection:
            await self.incoming_queue.put(message)

    async def producer(self):
        while True:
            message = await self.outgoing_queue.get()
            await self.connection.send(message.to_channels())

            self.outgoing_queue.task_done()

    async def forward(self, message: MessageModel):
        await self.outgoing_queue.put(message)


    async def workers(self):
        while True:
            message_str = await self.incoming_queue.get()
            message = expandToMessage(json.loads(message_str))
            logger.info(f"Received Message {message}")
            await self.on_message(message)
            self.incoming_queue.task_done()


    


    
    





from bergen.queries.delayed.node import DETAIL_NODE_FR
from bergen.schema import Assignation, Node, Peasent, Template, Pod, Provision, Transcript, VartPod, Volunteer
from bergen.query import TypedGQL


NEGOTIATION_GQL = TypedGQL("""

  mutation Negotiate($clientType: ClientTypeInput!) {
  negotiate(clientType: $clientType) {
    timestamp
    extensions
    models {
        identifier
        point {
            id
            outward
            port
            type
        }
    }
    postman {
        type
        kwargs
    }
    provider {
        type
        kwargs
    }
    host {
        type
        kwargs
    }
  }
  }
""", Transcript)

 
# Peasent Constants

SERVE_GQL = TypedGQL("""
    mutation Serve($name: String!){
        serve(name: $name){
            id
            name
            
        }
    }
""", Peasent)



OFFER_GQL = TypedGQL("""
    mutation Offer($node: ID!, $params: GenericScalar!, $policy: GenericScalar!){
        offer(node: $node, params: $params, policy: $policy){
            id
            name
            policy 
        }
    }
""", Template)


ACCEPT_GQL = TypedGQL("""
    mutation Accept($template: ID!, $provision: String!){
        accept(template: $template, provision: $provision){
            id
            template {
                node {"""
                + DETAIL_NODE_FR +
                """

                }
            }
        }
    }
""", Pod)




PROVIDE_GQL = TypedGQL("""
        subscription Provide($reference: String!, $node: ID!, $selector: SelectorInput!){
            provide(node: $node , selector: $selector, reference: $reference){
                pod {
                    id
                    status
                }
                node {
                    name
                }
                reference
                status
                statusmessage

            }
        }      
    """, Provision)



ASSIGN_GQL = TypedGQL("""
        subscription Assignation($inputs: GenericScalar!, $pod: ID!, $reference: String!) {
            assign(inputs: $inputs, pod: $pod, reference: $reference){
                inputs
                outputs
                status
                statusmessage
            }
        }     
""", Assignation)


from bergen.auths.types import HerreConfig
from oauthlib.oauth2.rfc6749.clients.legacy_application import LegacyApplicationClient
from oauthlib.oauth2.rfc6749.clients.mobile_application import MobileApplicationClient
import requests
from requests_oauthlib.oauth2_session import OAuth2Session
from bergen.auths.base import BaseAuthBackend
from bergen.enums import ClientType
from bergen.console import console

class ImplicitError(Exception):
    pass


class LegacyApplication(BaseAuthBackend):


    def __init__(self, config: HerreConfig, username=None, password=None, **kwargs) -> None:
        super().__init__(config, **kwargs)
        self.username = username
        self.password = password

        self.oauth = OAuth2Session(client=LegacyApplicationClient(client_id=self.client_id))

    
    def fetchToken(self, loop=None) -> str:
        # Getting token
        self.legacy_app_client =  LegacyApplicationClient(self.client_id, scope=self.scope)
        console.print("[yellow]Try to avoid authenticating with User and Password. Use Implicit Flow if possible!")
        if not self.username: self.username = console.input("[green]Enter your [b]Username:               ")
        if not self.password: self.password = console.input("[green]Now for the [b]Password:              ", password=True)

        with console.status("[bold green]Authenticating..."):
            token = self.oauth.fetch_token(token_url=self.token_url,
            username=self.username, password=self.password, client_id=self.client_id,
            client_secret=self.client_secret)
        
        return token



import logging

def setLogging(log_level, log_stream=False):
    LOGFORMAT = "  s%(levelname)-8s%(reset)s | %(message)s%(reset)s"
    file = logging.FileHandler("logs.txt")
    logging.root.setLevel(log_level)
    file.setLevel(log_level)


    
    logger = logging.getLogger()
    logger.setLevel(log_level)

    if log_stream:
        stream = logging.StreamHandler()
        logging.root.setLevel(log_level)
        stream.setLevel(log_level)

        try:
            from colorlog import ColoredFormatter
            LOGFORMAT = "  %(log_color)s%(levelname)-8s%(reset)s | %(message)s%(reset)s"
            formatter = ColoredFormatter(LOGFORMAT)
            stream.setFormatter(formatter)
        except:
            pass

        logger.addHandler(stream)

    logger.addHandler(file)



from typing import Any
from bergen.messages.postman.reserve.reserve_critical import ReserveCriticalMessage
from bergen.schema import Template
from bergen.messages.postman.assign.assign_done import AssignDoneMessage
from bergen.debugging import DebugLevel
from bergen.handlers.base import Connector, ContractHandler
from bergen.messages import BouncedForwardedReserveMessage
from bergen.console import console


class ReserveHandler(ContractHandler[BouncedForwardedReserveMessage]):

    def __init__(self, message: BouncedForwardedReserveMessage, connection: Connector) -> None:
        super().__init__(message, connection)
        self.active_context = None
        self.context_set = False

    @property
    def bounced(self):
        return self.message.meta.token

    @property
    def meta(self):
        return {"extensions": self.message.meta.extensions, "reference": self.message.meta.reference}

    @property
    def template_id(self) -> str:
        return self.message.data.template

    @property
    def context(self) -> Any:
        assert self.context_set, "Context was never set, no sense in accessing it..."
        return self.active_context   


    def set_context(self, context):
        self.context_set = True
        self.active_context = context

    async def get_template(self) -> Template:
        return await Template.asyncs.get(id=self.message.data.template)

    async def pass_done(self):
        return await self.connection.forward(self.message) #Contrary to a Done signal we are the provision just to the backend, because it will take care of it

    async def pass_exception(self, exception):
        critical_message = ReserveCriticalMessage(data={"message": str(exception), "type": exception.__class__.__name__}, meta=self.meta)
        console.print_exception()
        return await self.connection.forward(critical_message)



from bergen.schema import Template
from bergen.messages.postman.assign.assign_done import AssignDoneMessage
from bergen.debugging import DebugLevel
from bergen.handlers.base import ContractHandler
from bergen.messages import BouncedProvideMessage
from bergen.console import console


class ProvideHandler(ContractHandler[BouncedProvideMessage]):

    @property
    def template_id(self) -> str:
        return self.message.data.template

    async def get_template(self) -> Template:
        return await Template.asyncs.get(id=self.message.data.template)


    async def pass_exception(self, exception):
        console.print_exception()